// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.6
// 	protoc        v3.21.12
// source: agentapi.proto

package agentapi

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Empty struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Empty) Reset() {
	*x = Empty{}
	mi := &file_agentapi_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Empty) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Empty) ProtoMessage() {}

func (x *Empty) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Empty.ProtoReflect.Descriptor instead.
func (*Empty) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{0}
}

type ProAttachInfo struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Token         string                 `protobuf:"bytes,1,opt,name=token,proto3" json:"token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProAttachInfo) Reset() {
	*x = ProAttachInfo{}
	mi := &file_agentapi_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProAttachInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProAttachInfo) ProtoMessage() {}

func (x *ProAttachInfo) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProAttachInfo.ProtoReflect.Descriptor instead.
func (*ProAttachInfo) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{1}
}

func (x *ProAttachInfo) GetToken() string {
	if x != nil {
		return x.Token
	}
	return ""
}

type LandscapeConfig struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Config        string                 `protobuf:"bytes,1,opt,name=config,proto3" json:"config,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *LandscapeConfig) Reset() {
	*x = LandscapeConfig{}
	mi := &file_agentapi_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *LandscapeConfig) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*LandscapeConfig) ProtoMessage() {}

func (x *LandscapeConfig) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use LandscapeConfig.ProtoReflect.Descriptor instead.
func (*LandscapeConfig) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{2}
}

func (x *LandscapeConfig) GetConfig() string {
	if x != nil {
		return x.Config
	}
	return ""
}

type SubscriptionInfo struct {
	state     protoimpl.MessageState `protogen:"open.v1"`
	ProductId string                 `protobuf:"bytes,1,opt,name=productId,proto3" json:"productId,omitempty"` // The ID of the Ubuntu Pro for WSL product on the Microsoft Store.
	// Types that are valid to be assigned to SubscriptionType:
	//
	//	*SubscriptionInfo_None
	//	*SubscriptionInfo_User
	//	*SubscriptionInfo_Organization
	//	*SubscriptionInfo_MicrosoftStore
	SubscriptionType isSubscriptionInfo_SubscriptionType `protobuf_oneof:"subscriptionType"`
	unknownFields    protoimpl.UnknownFields
	sizeCache        protoimpl.SizeCache
}

func (x *SubscriptionInfo) Reset() {
	*x = SubscriptionInfo{}
	mi := &file_agentapi_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SubscriptionInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SubscriptionInfo) ProtoMessage() {}

func (x *SubscriptionInfo) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SubscriptionInfo.ProtoReflect.Descriptor instead.
func (*SubscriptionInfo) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{3}
}

func (x *SubscriptionInfo) GetProductId() string {
	if x != nil {
		return x.ProductId
	}
	return ""
}

func (x *SubscriptionInfo) GetSubscriptionType() isSubscriptionInfo_SubscriptionType {
	if x != nil {
		return x.SubscriptionType
	}
	return nil
}

func (x *SubscriptionInfo) GetNone() *Empty {
	if x != nil {
		if x, ok := x.SubscriptionType.(*SubscriptionInfo_None); ok {
			return x.None
		}
	}
	return nil
}

func (x *SubscriptionInfo) GetUser() *Empty {
	if x != nil {
		if x, ok := x.SubscriptionType.(*SubscriptionInfo_User); ok {
			return x.User
		}
	}
	return nil
}

func (x *SubscriptionInfo) GetOrganization() *Empty {
	if x != nil {
		if x, ok := x.SubscriptionType.(*SubscriptionInfo_Organization); ok {
			return x.Organization
		}
	}
	return nil
}

func (x *SubscriptionInfo) GetMicrosoftStore() *Empty {
	if x != nil {
		if x, ok := x.SubscriptionType.(*SubscriptionInfo_MicrosoftStore); ok {
			return x.MicrosoftStore
		}
	}
	return nil
}

type isSubscriptionInfo_SubscriptionType interface {
	isSubscriptionInfo_SubscriptionType()
}

type SubscriptionInfo_None struct {
	None *Empty `protobuf:"bytes,2,opt,name=none,proto3,oneof"` // There is no active subscription.
}

type SubscriptionInfo_User struct {
	User *Empty `protobuf:"bytes,3,opt,name=user,proto3,oneof"` // The subscription is managed by the user with a pro token from the GUI or the registry.
}

type SubscriptionInfo_Organization struct {
	Organization *Empty `protobuf:"bytes,4,opt,name=organization,proto3,oneof"` // The subscription is managed by the sysadmin with a pro token from the registry.
}

type SubscriptionInfo_MicrosoftStore struct {
	MicrosoftStore *Empty `protobuf:"bytes,5,opt,name=microsoftStore,proto3,oneof"` // The subscription is managed via the Microsoft store.
}

func (*SubscriptionInfo_None) isSubscriptionInfo_SubscriptionType() {}

func (*SubscriptionInfo_User) isSubscriptionInfo_SubscriptionType() {}

func (*SubscriptionInfo_Organization) isSubscriptionInfo_SubscriptionType() {}

func (*SubscriptionInfo_MicrosoftStore) isSubscriptionInfo_SubscriptionType() {}

type LandscapeSource struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Types that are valid to be assigned to LandscapeSourceType:
	//
	//	*LandscapeSource_None
	//	*LandscapeSource_User
	//	*LandscapeSource_Organization
	LandscapeSourceType isLandscapeSource_LandscapeSourceType `protobuf_oneof:"landscapeSourceType"`
	unknownFields       protoimpl.UnknownFields
	sizeCache           protoimpl.SizeCache
}

func (x *LandscapeSource) Reset() {
	*x = LandscapeSource{}
	mi := &file_agentapi_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *LandscapeSource) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*LandscapeSource) ProtoMessage() {}

func (x *LandscapeSource) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use LandscapeSource.ProtoReflect.Descriptor instead.
func (*LandscapeSource) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{4}
}

func (x *LandscapeSource) GetLandscapeSourceType() isLandscapeSource_LandscapeSourceType {
	if x != nil {
		return x.LandscapeSourceType
	}
	return nil
}

func (x *LandscapeSource) GetNone() *Empty {
	if x != nil {
		if x, ok := x.LandscapeSourceType.(*LandscapeSource_None); ok {
			return x.None
		}
	}
	return nil
}

func (x *LandscapeSource) GetUser() *Empty {
	if x != nil {
		if x, ok := x.LandscapeSourceType.(*LandscapeSource_User); ok {
			return x.User
		}
	}
	return nil
}

func (x *LandscapeSource) GetOrganization() *Empty {
	if x != nil {
		if x, ok := x.LandscapeSourceType.(*LandscapeSource_Organization); ok {
			return x.Organization
		}
	}
	return nil
}

type isLandscapeSource_LandscapeSourceType interface {
	isLandscapeSource_LandscapeSourceType()
}

type LandscapeSource_None struct {
	None *Empty `protobuf:"bytes,1,opt,name=none,proto3,oneof"` // There is no active Landscape config data.
}

type LandscapeSource_User struct {
	User *Empty `protobuf:"bytes,2,opt,name=user,proto3,oneof"` // The Landscape config is managed by the user, set via the GUI.
}

type LandscapeSource_Organization struct {
	Organization *Empty `protobuf:"bytes,3,opt,name=organization,proto3,oneof"` // The Landscape config is managedby the sysadmin, set via the registry.
}

func (*LandscapeSource_None) isLandscapeSource_LandscapeSourceType() {}

func (*LandscapeSource_User) isLandscapeSource_LandscapeSourceType() {}

func (*LandscapeSource_Organization) isLandscapeSource_LandscapeSourceType() {}

type ConfigSources struct {
	state           protoimpl.MessageState `protogen:"open.v1"`
	ProSubscription *SubscriptionInfo      `protobuf:"bytes,1,opt,name=proSubscription,proto3" json:"proSubscription,omitempty"`
	LandscapeSource *LandscapeSource       `protobuf:"bytes,2,opt,name=landscapeSource,proto3" json:"landscapeSource,omitempty"`
	unknownFields   protoimpl.UnknownFields
	sizeCache       protoimpl.SizeCache
}

func (x *ConfigSources) Reset() {
	*x = ConfigSources{}
	mi := &file_agentapi_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ConfigSources) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ConfigSources) ProtoMessage() {}

func (x *ConfigSources) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ConfigSources.ProtoReflect.Descriptor instead.
func (*ConfigSources) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{5}
}

func (x *ConfigSources) GetProSubscription() *SubscriptionInfo {
	if x != nil {
		return x.ProSubscription
	}
	return nil
}

func (x *ConfigSources) GetLandscapeSource() *LandscapeSource {
	if x != nil {
		return x.LandscapeSource
	}
	return nil
}

type DistroInfo struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	WslName       string                 `protobuf:"bytes,1,opt,name=wsl_name,json=wslName,proto3" json:"wsl_name,omitempty"`
	Id            string                 `protobuf:"bytes,2,opt,name=id,proto3" json:"id,omitempty"`
	VersionId     string                 `protobuf:"bytes,3,opt,name=version_id,json=versionId,proto3" json:"version_id,omitempty"`
	PrettyName    string                 `protobuf:"bytes,4,opt,name=pretty_name,json=prettyName,proto3" json:"pretty_name,omitempty"`
	ProAttached   bool                   `protobuf:"varint,5,opt,name=pro_attached,json=proAttached,proto3" json:"pro_attached,omitempty"`
	Hostname      string                 `protobuf:"bytes,6,opt,name=hostname,proto3" json:"hostname,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DistroInfo) Reset() {
	*x = DistroInfo{}
	mi := &file_agentapi_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DistroInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DistroInfo) ProtoMessage() {}

func (x *DistroInfo) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DistroInfo.ProtoReflect.Descriptor instead.
func (*DistroInfo) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{6}
}

func (x *DistroInfo) GetWslName() string {
	if x != nil {
		return x.WslName
	}
	return ""
}

func (x *DistroInfo) GetId() string {
	if x != nil {
		return x.Id
	}
	return ""
}

func (x *DistroInfo) GetVersionId() string {
	if x != nil {
		return x.VersionId
	}
	return ""
}

func (x *DistroInfo) GetPrettyName() string {
	if x != nil {
		return x.PrettyName
	}
	return ""
}

func (x *DistroInfo) GetProAttached() bool {
	if x != nil {
		return x.ProAttached
	}
	return false
}

func (x *DistroInfo) GetHostname() string {
	if x != nil {
		return x.Hostname
	}
	return ""
}

type ProAttachCmd struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Token         string                 `protobuf:"bytes,1,opt,name=token,proto3" json:"token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProAttachCmd) Reset() {
	*x = ProAttachCmd{}
	mi := &file_agentapi_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProAttachCmd) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProAttachCmd) ProtoMessage() {}

func (x *ProAttachCmd) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProAttachCmd.ProtoReflect.Descriptor instead.
func (*ProAttachCmd) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{7}
}

func (x *ProAttachCmd) GetToken() string {
	if x != nil {
		return x.Token
	}
	return ""
}

type LandscapeConfigCmd struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Config        string                 `protobuf:"bytes,1,opt,name=config,proto3" json:"config,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *LandscapeConfigCmd) Reset() {
	*x = LandscapeConfigCmd{}
	mi := &file_agentapi_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *LandscapeConfigCmd) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*LandscapeConfigCmd) ProtoMessage() {}

func (x *LandscapeConfigCmd) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use LandscapeConfigCmd.ProtoReflect.Descriptor instead.
func (*LandscapeConfigCmd) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{8}
}

func (x *LandscapeConfigCmd) GetConfig() string {
	if x != nil {
		return x.Config
	}
	return ""
}

type MSG struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Types that are valid to be assigned to Data:
	//
	//	*MSG_WslName
	//	*MSG_Result
	Data          isMSG_Data `protobuf_oneof:"data"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *MSG) Reset() {
	*x = MSG{}
	mi := &file_agentapi_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *MSG) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*MSG) ProtoMessage() {}

func (x *MSG) ProtoReflect() protoreflect.Message {
	mi := &file_agentapi_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use MSG.ProtoReflect.Descriptor instead.
func (*MSG) Descriptor() ([]byte, []int) {
	return file_agentapi_proto_rawDescGZIP(), []int{9}
}

func (x *MSG) GetData() isMSG_Data {
	if x != nil {
		return x.Data
	}
	return nil
}

func (x *MSG) GetWslName() string {
	if x != nil {
		if x, ok := x.Data.(*MSG_WslName); ok {
			return x.WslName
		}
	}
	return ""
}

func (x *MSG) GetResult() string {
	if x != nil {
		if x, ok := x.Data.(*MSG_Result); ok {
			return x.Result
		}
	}
	return ""
}

type isMSG_Data interface {
	isMSG_Data()
}

type MSG_WslName struct {
	WslName string `protobuf:"bytes,1,opt,name=wsl_name,json=wslName,proto3,oneof"` // Used during handshake to identify the WSL instance.
}

type MSG_Result struct {
	Result string `protobuf:"bytes,2,opt,name=result,proto3,oneof"` // Used in response to a command
}

func (*MSG_WslName) isMSG_Data() {}

func (*MSG_Result) isMSG_Data() {}

var File_agentapi_proto protoreflect.FileDescriptor

const file_agentapi_proto_rawDesc = "" +
	"\n" +
	"\x0eagentapi.proto\x12\bagentapi\"\a\n" +
	"\x05Empty\"%\n" +
	"\rProAttachInfo\x12\x14\n" +
	"\x05token\x18\x01 \x01(\tR\x05token\")\n" +
	"\x0fLandscapeConfig\x12\x16\n" +
	"\x06config\x18\x01 \x01(\tR\x06config\"\x84\x02\n" +
	"\x10SubscriptionInfo\x12\x1c\n" +
	"\tproductId\x18\x01 \x01(\tR\tproductId\x12%\n" +
	"\x04none\x18\x02 \x01(\v2\x0f.agentapi.EmptyH\x00R\x04none\x12%\n" +
	"\x04user\x18\x03 \x01(\v2\x0f.agentapi.EmptyH\x00R\x04user\x125\n" +
	"\forganization\x18\x04 \x01(\v2\x0f.agentapi.EmptyH\x00R\forganization\x129\n" +
	"\x0emicrosoftStore\x18\x05 \x01(\v2\x0f.agentapi.EmptyH\x00R\x0emicrosoftStoreB\x12\n" +
	"\x10subscriptionType\"\xad\x01\n" +
	"\x0fLandscapeSource\x12%\n" +
	"\x04none\x18\x01 \x01(\v2\x0f.agentapi.EmptyH\x00R\x04none\x12%\n" +
	"\x04user\x18\x02 \x01(\v2\x0f.agentapi.EmptyH\x00R\x04user\x125\n" +
	"\forganization\x18\x03 \x01(\v2\x0f.agentapi.EmptyH\x00R\forganizationB\x15\n" +
	"\x13landscapeSourceType\"\x9a\x01\n" +
	"\rConfigSources\x12D\n" +
	"\x0fproSubscription\x18\x01 \x01(\v2\x1a.agentapi.SubscriptionInfoR\x0fproSubscription\x12C\n" +
	"\x0flandscapeSource\x18\x02 \x01(\v2\x19.agentapi.LandscapeSourceR\x0flandscapeSource\"\xb6\x01\n" +
	"\n" +
	"DistroInfo\x12\x19\n" +
	"\bwsl_name\x18\x01 \x01(\tR\awslName\x12\x0e\n" +
	"\x02id\x18\x02 \x01(\tR\x02id\x12\x1d\n" +
	"\n" +
	"version_id\x18\x03 \x01(\tR\tversionId\x12\x1f\n" +
	"\vpretty_name\x18\x04 \x01(\tR\n" +
	"prettyName\x12!\n" +
	"\fpro_attached\x18\x05 \x01(\bR\vproAttached\x12\x1a\n" +
	"\bhostname\x18\x06 \x01(\tR\bhostname\"$\n" +
	"\fProAttachCmd\x12\x14\n" +
	"\x05token\x18\x01 \x01(\tR\x05token\",\n" +
	"\x12LandscapeConfigCmd\x12\x16\n" +
	"\x06config\x18\x01 \x01(\tR\x06config\"D\n" +
	"\x03MSG\x12\x1b\n" +
	"\bwsl_name\x18\x01 \x01(\tH\x00R\awslName\x12\x18\n" +
	"\x06result\x18\x02 \x01(\tH\x00R\x06resultB\x06\n" +
	"\x04data2\xc9\x02\n" +
	"\x02UI\x12F\n" +
	"\rApplyProToken\x12\x17.agentapi.ProAttachInfo\x1a\x1a.agentapi.SubscriptionInfo\"\x00\x12N\n" +
	"\x14ApplyLandscapeConfig\x12\x19.agentapi.LandscapeConfig\x1a\x19.agentapi.LandscapeSource\"\x00\x12*\n" +
	"\x04Ping\x12\x0f.agentapi.Empty\x1a\x0f.agentapi.Empty\"\x00\x12>\n" +
	"\x10GetConfigSources\x12\x0f.agentapi.Empty\x1a\x17.agentapi.ConfigSources\"\x00\x12?\n" +
	"\x0eNotifyPurchase\x12\x0f.agentapi.Empty\x1a\x1a.agentapi.SubscriptionInfo\"\x002\xd9\x01\n" +
	"\vWSLInstance\x126\n" +
	"\tConnected\x12\x14.agentapi.DistroInfo\x1a\x0f.agentapi.Empty\"\x00(\x01\x12D\n" +
	"\x15ProAttachmentCommands\x12\r.agentapi.MSG\x1a\x16.agentapi.ProAttachCmd\"\x00(\x010\x01\x12L\n" +
	"\x17LandscapeConfigCommands\x12\r.agentapi.MSG\x1a\x1c.agentapi.LandscapeConfigCmd\"\x00(\x010\x01B2Z0github.com/canonical/ubuntu-pro-for-wsl/agentapib\x06proto3"

var (
	file_agentapi_proto_rawDescOnce sync.Once
	file_agentapi_proto_rawDescData []byte
)

func file_agentapi_proto_rawDescGZIP() []byte {
	file_agentapi_proto_rawDescOnce.Do(func() {
		file_agentapi_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_agentapi_proto_rawDesc), len(file_agentapi_proto_rawDesc)))
	})
	return file_agentapi_proto_rawDescData
}

var file_agentapi_proto_msgTypes = make([]protoimpl.MessageInfo, 10)
var file_agentapi_proto_goTypes = []any{
	(*Empty)(nil),              // 0: agentapi.Empty
	(*ProAttachInfo)(nil),      // 1: agentapi.ProAttachInfo
	(*LandscapeConfig)(nil),    // 2: agentapi.LandscapeConfig
	(*SubscriptionInfo)(nil),   // 3: agentapi.SubscriptionInfo
	(*LandscapeSource)(nil),    // 4: agentapi.LandscapeSource
	(*ConfigSources)(nil),      // 5: agentapi.ConfigSources
	(*DistroInfo)(nil),         // 6: agentapi.DistroInfo
	(*ProAttachCmd)(nil),       // 7: agentapi.ProAttachCmd
	(*LandscapeConfigCmd)(nil), // 8: agentapi.LandscapeConfigCmd
	(*MSG)(nil),                // 9: agentapi.MSG
}
var file_agentapi_proto_depIdxs = []int32{
	0,  // 0: agentapi.SubscriptionInfo.none:type_name -> agentapi.Empty
	0,  // 1: agentapi.SubscriptionInfo.user:type_name -> agentapi.Empty
	0,  // 2: agentapi.SubscriptionInfo.organization:type_name -> agentapi.Empty
	0,  // 3: agentapi.SubscriptionInfo.microsoftStore:type_name -> agentapi.Empty
	0,  // 4: agentapi.LandscapeSource.none:type_name -> agentapi.Empty
	0,  // 5: agentapi.LandscapeSource.user:type_name -> agentapi.Empty
	0,  // 6: agentapi.LandscapeSource.organization:type_name -> agentapi.Empty
	3,  // 7: agentapi.ConfigSources.proSubscription:type_name -> agentapi.SubscriptionInfo
	4,  // 8: agentapi.ConfigSources.landscapeSource:type_name -> agentapi.LandscapeSource
	1,  // 9: agentapi.UI.ApplyProToken:input_type -> agentapi.ProAttachInfo
	2,  // 10: agentapi.UI.ApplyLandscapeConfig:input_type -> agentapi.LandscapeConfig
	0,  // 11: agentapi.UI.Ping:input_type -> agentapi.Empty
	0,  // 12: agentapi.UI.GetConfigSources:input_type -> agentapi.Empty
	0,  // 13: agentapi.UI.NotifyPurchase:input_type -> agentapi.Empty
	6,  // 14: agentapi.WSLInstance.Connected:input_type -> agentapi.DistroInfo
	9,  // 15: agentapi.WSLInstance.ProAttachmentCommands:input_type -> agentapi.MSG
	9,  // 16: agentapi.WSLInstance.LandscapeConfigCommands:input_type -> agentapi.MSG
	3,  // 17: agentapi.UI.ApplyProToken:output_type -> agentapi.SubscriptionInfo
	4,  // 18: agentapi.UI.ApplyLandscapeConfig:output_type -> agentapi.LandscapeSource
	0,  // 19: agentapi.UI.Ping:output_type -> agentapi.Empty
	5,  // 20: agentapi.UI.GetConfigSources:output_type -> agentapi.ConfigSources
	3,  // 21: agentapi.UI.NotifyPurchase:output_type -> agentapi.SubscriptionInfo
	0,  // 22: agentapi.WSLInstance.Connected:output_type -> agentapi.Empty
	7,  // 23: agentapi.WSLInstance.ProAttachmentCommands:output_type -> agentapi.ProAttachCmd
	8,  // 24: agentapi.WSLInstance.LandscapeConfigCommands:output_type -> agentapi.LandscapeConfigCmd
	17, // [17:25] is the sub-list for method output_type
	9,  // [9:17] is the sub-list for method input_type
	9,  // [9:9] is the sub-list for extension type_name
	9,  // [9:9] is the sub-list for extension extendee
	0,  // [0:9] is the sub-list for field type_name
}

func init() { file_agentapi_proto_init() }
func file_agentapi_proto_init() {
	if File_agentapi_proto != nil {
		return
	}
	file_agentapi_proto_msgTypes[3].OneofWrappers = []any{
		(*SubscriptionInfo_None)(nil),
		(*SubscriptionInfo_User)(nil),
		(*SubscriptionInfo_Organization)(nil),
		(*SubscriptionInfo_MicrosoftStore)(nil),
	}
	file_agentapi_proto_msgTypes[4].OneofWrappers = []any{
		(*LandscapeSource_None)(nil),
		(*LandscapeSource_User)(nil),
		(*LandscapeSource_Organization)(nil),
	}
	file_agentapi_proto_msgTypes[9].OneofWrappers = []any{
		(*MSG_WslName)(nil),
		(*MSG_Result)(nil),
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_agentapi_proto_rawDesc), len(file_agentapi_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   10,
			NumExtensions: 0,
			NumServices:   2,
		},
		GoTypes:           file_agentapi_proto_goTypes,
		DependencyIndexes: file_agentapi_proto_depIdxs,
		MessageInfos:      file_agentapi_proto_msgTypes,
	}.Build()
	File_agentapi_proto = out.File
	file_agentapi_proto_goTypes = nil
	file_agentapi_proto_depIdxs = nil
}
